# -*- coding: utf-8 -*-

__version__ = "0.2.2"

__title__ = "pygcgen"
__summary__ = "Automatic changelog generation"
__uri__ = "https://github.com/topic2k/pygcgen"

__author__ = "topic2k"
__email__ = "topic2k+pypi@gmail.com"

__license__ = "MIT"
__copyright__ = "2016 %s" % __author__
